// #define the implementations

#define DRFLAC_IMPLEMENTATION
#include "dr_flac.h"

#define MINIAUDIO_IMPLEMENTATION
#include "miniaudio.h"

#define BLOCKVF_IMPLEMENTATION
#include "blockvf.h"

#define MHFSCL_IMPLEMENTATION
#include "mhfs_cl.h"

#define MHFSCLTRACK_IMPLEMENTATION
#include "mhfs_cl_track.h"

#define MHFSCLDECODER_IMPLEMENTATION
#include "mhfs_cl_decoder.h"

#define MHFSCLMISC_IMPLEMENTATION
#include "mhfs_cl_misc.h"
